# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
# Authors: Amador Pahim <amador@pahim.org>
#          Bestoun S. Ahmed <bestoon82@gmail.com>
#          Cleber Rosa <crosa@redhat.com>

import logging
import os
import sys

from avocado_varianter_cit.Cit import LOG, Cit
from avocado_varianter_cit.Parser import Parser

from avocado.core import exit_codes, varianter
from avocado.core.output import LOG_UI
from avocado.core.plugin_interfaces import CLI, Varianter
from avocado.core.settings import settings
from avocado.core.tree import TreeNode

#: The default order of combinations
DEFAULT_ORDER_OF_COMBINATIONS = 2


class VarianterCitCLI(CLI):

    """
    CIT Varianter options
    """

    name = "cit"
    description = "CIT Varianter options for the 'run' subcommand"

    def configure(self, parser):

        for name in ("run", "variants"):
            subparser = parser.subcommands.choices.get(name, None)
            if subparser is None:
                continue
            subparser.add_argument_group("CIT varianter options")
            settings.register_option(
                section=f"{name}.cit",
                key="parameter_file",
                metavar="PATH",
                help_msg="Paths to a parameter file",
                parser=subparser,
                default=None,
                long_arg="--cit-parameter-file",
            )

            help_msg = "Order of combinations. Maximum number is 6"
            settings.register_option(
                section=f"{name}.cit",
                key="combination_order",
                key_type=int,
                parser=subparser,
                help_msg=help_msg,
                metavar="ORDER",
                default=DEFAULT_ORDER_OF_COMBINATIONS,
                long_arg="--cit-order-of-combinations",
            )

    def run(self, config):
        if config.get("variants.debug"):
            LOG.setLevel(logging.DEBUG)


class VarianterCit(Varianter):

    """
    Processes the parameters file into variants
    """

    name = "cit"
    description = "CIT Varianter"

    def initialize(self, config):
        subcommand = config.get("subcommand")
        self.variants = None  # pylint: disable=W0201
        order = config.get(f"{subcommand}.cit.combination_order")
        if order and order > 6:
            LOG_UI.error("The order of combinations is bigger then 6")
            self.error_exit(config)

        section_key = f"{subcommand}.cit.parameter_file"
        cit_parameter_file = config.get(section_key)
        if cit_parameter_file is None:
            return
        else:
            cit_parameter_file = os.path.expanduser(cit_parameter_file)
            if not os.access(cit_parameter_file, os.R_OK):
                LOG_UI.error(
                    "parameter file '%s' could not be found or " "is not readable",
                    cit_parameter_file,
                )
                self.error_exit(config)

        try:
            parameters, constraints = Parser.parse(
                open(cit_parameter_file, encoding="utf-8")
            )
        except ValueError as details:
            LOG_UI.error("Cannot parse parameter file: %s", details)
            self.error_exit(config)

        input_data = [len(parameter[1]) for parameter in parameters]

        cit = Cit(input_data, order, constraints)
        final_list = cit.compute()
        self.headers = [  # pylint: disable=W0201
            parameter[0] for parameter in parameters
        ]
        results = [
            [parameters[j][1][final_list[i][j]] for j in range(len(final_list[i]))]
            for i in range(len(final_list))
        ]
        self.variants = []  # pylint: disable=W0201
        for combination in results:
            self.variants.append(dict(zip(self.headers, combination)))

    @staticmethod
    def error_exit(config):
        if config.get("subcommand") == "run":
            sys.exit(exit_codes.AVOCADO_JOB_FAIL)
        else:
            sys.exit(exit_codes.AVOCADO_FAIL)

    def __iter__(self):
        if self.variants is None:
            return

        variant_ids = []
        for variant in self.variants:
            variant_ids.append("-".join([variant.get(key) for key in self.headers]))

        for vid, variant in zip(variant_ids, self.variants):
            yield {
                "variant_id": vid,
                "variant": [TreeNode("", variant)],
                "paths": ["/"],
            }

    def __len__(self):
        return sum(1 for _ in self.variants) if self.variants else 0

    def to_str(self, summary, variants, **kwargs):
        """
        Return human readable representation

        The summary/variants accepts verbosity where 0 means silent and
        maximum is up to the plugin.

        :param summary: How verbose summary to output (int)
        :param variants: How verbose list of variants to output (int)
        :param kwargs: Other free-form arguments
        :rtype: str
        """
        if not self.variants:
            return ""
        out = []

        if variants:
            # variants == 0 means disable, but in plugin it's brief
            out.append(f"CIT Variants ({len(self)}):")
            for variant in self:
                out.extend(
                    varianter.variant_to_str(variant, variants - 1, kwargs, False)
                )
        return "\n".join(out)



import itertools


class Parameter:
    """
    Storage for constraints of one parameter.

    This class stores the constraints which constrain the values of
    one parameter.

    :param name: identification of parameter
    :type name: int
    :param size: number of values
    :type size: int
    :param constraints: list for storing constraints
    :type constraints: list
    """

    def __init__(self, name, values):
        """
        Parameter initialization.

        :param name:  identification of parameter
        :type name: int
        :param values: values of parameter
        :type: list
        """
        self.name = name
        self.size = len(values)
        self._constrained_values_number = 0
        self.constraints = [None] * self.size

    def add_constraint(self, constraint, value, index):
        """
        Append new constraint to the parameter.

        The constraint is placed under the parameter value which is affected by
        this constraint. And this value is also deleted from the constraint,
        because is defined by the index in the 'self.constraints' list.

        :param constraint: will be appended to the parameter constraints
        :type constraint: list
        :param value: parameter value which is is affected by new constraint
        :type value: int
        :param index: index of that value inside the constraint
        :type index: int
        """
        if self.constraints[value] is None:
            self._constrained_values_number += 1
            self.constraints[value] = []
        array = list(constraint)
        array.pop(index)
        if len(array) != 0:
            self.constraints[value].append(array)

    @property
    def is_full(self):
        """
        Compute if constraints constrain every parameter value.

        :rtype: bool
        """
        return self._constrained_values_number == self.size


class Solver:

    CON_NAME = 0
    CON_VAL = 1

    def __init__(self, data, constraints):
        self.data = data
        self.constraints = constraints
        self.parameters = []

        self.simplify_constraints()
        constraint_size = len(self.constraints)
        self.read_constraints()
        self.compute_constraints()
        self.simplify_constraints()
        while constraint_size != len(self.constraints):
            constraint_size = len(self.constraints)
            self.parameters = []
            self.read_constraints()
            self.compute_constraints()
            self.simplify_constraints()

    def read_constraints(self):
        # creates new parameters with their names
        for i, values_size in enumerate(self.data):
            self.parameters.append(Parameter(i, list(range(values_size))))
        for constraint in self.constraints:
            for index, pair in enumerate(constraint):
                self.parameters[pair[self.CON_NAME]].add_constraint(
                    constraint, pair[self.CON_VAL], index
                )

    def compute_constraints(self):
        for p in self.parameters:
            if p.is_full:
                array = [c for c in p.constraints if len(c) != 0]
                con = list(itertools.product(*array))
                if len(con[0]) == 0:
                    raise ValueError("Constraints are not satisfiable")
                for constraint in con:
                    constraint_array = set()
                    for c in range(len(constraint)):
                        for pair in range(len(constraint[c])):
                            constraint_array.add(constraint[c][pair])
                    constraint_array = sorted(
                        constraint_array, key=lambda x: int(x[self.CON_NAME])
                    )

                    has_subset = False
                    remove = set()
                    for c in self.constraints:
                        if len(c) < len(constraint_array):
                            if set(c) < set(constraint_array):
                                has_subset = True
                                break
                        if len(c) > len(constraint_array):
                            if set(c) > set(constraint_array):
                                remove.add(c)
                    if not has_subset:
                        self.constraints.add(tuple(constraint_array))
                    for r in remove:
                        self.constraints.remove(r)

    def simplify_constraints(self):
        items_to_remove = set()
        copy = list(self.constraints.copy())
        for i in range(len(copy)):
            is_brake = False
            for j in range(len(copy[i])):
                for k in range(j + 1, len(copy[i])):
                    if copy[i][j][self.CON_NAME] == copy[i][k][self.CON_NAME]:
                        items_to_remove.add(copy[i])
                        is_brake = True
                        break
                if is_brake:
                    break
            if is_brake:
                continue
            for j in range(len(copy)):
                if j != i:
                    if len(copy[i]) < len(copy[j]):
                        if set(copy[i]).issubset(set(copy[j])):
                            items_to_remove.add(copy[j])
        for item in items_to_remove:
            self.constraints.remove(item)

    def clean_hash_table(self, combination_matrix, t_value):
        for constraint in self.constraints:
            if len(constraint) > t_value:
                continue
            parameters_in_constraint = []
            for pair in constraint:
                parameters_in_constraint.append(pair[self.CON_NAME])
            for c in itertools.combinations(range(len(self.data)), t_value):
                if set(parameters_in_constraint).issubset(c):
                    value_array = []
                    counter = 0
                    for value in c:
                        if value == constraint[counter][self.CON_NAME]:
                            value_array.append([constraint[counter][self.CON_VAL]])
                            if (counter + 1) != len(constraint):
                                counter += 1
                        else:
                            value_array.append(list(range(0, self.data[value])))
                    for key in itertools.product(*value_array):
                        combination_matrix.del_cell(c, key)

    def get_possible_values(self, row, parameter):
        """
        Compute all possible values for the given parameter.

        These values are based on constraints and already picked values
        of other parameters.

        :param row: row with picked values. -1 means an unpicked value.
        :type row: list
        :param parameter: index of the parameter
         whose we want to know the values
        :type parameter: int
        :return: all possible values for the given parameter
        :rtype: list
        """

        def is_permitted_value(one_value_constraints):
            if one_value_constraints is None:
                return True
            if len(one_value_constraints) == 0:
                return False

            for constraints in one_value_constraints:
                is_ok = False
                for constraint in constraints:
                    if row[constraint[self.CON_NAME]] != constraint[self.CON_VAL]:
                        is_ok = True
                        break
                if not is_ok:
                    return False
            return True

        return [
            i
            for i, c in enumerate(self.parameters[parameter].constraints)
            if is_permitted_value(c)
        ]



#!/bin/env python3
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
# Authors: Amador Pahim <amador@pahim.org>
#          Bestoun S. Ahmed <bestoon82@gmail.com>
#          Cleber Rosa <crosa@redhat.com>


from setuptools import setup

# Handle systems with setuptools < 40
try:
    from setuptools import find_namespace_packages
except ImportError:
    packages = ["avocado_varianter_cit"]
else:
    packages = find_namespace_packages(include=["avocado_varianter_cit"])

VERSION = open("VERSION", "r", encoding="utf-8").read().strip()

setup(
    name="avocado-framework-plugin-varianter-cit",
    description="Varianter with combinatorial capabilities",
    version=open("VERSION", "r", encoding="utf-8").read().strip(),
    author="Avocado Developers",
    author_email="avocado-devel@redhat.com",
    url="http://avocado-framework.github.io/",
    packages=packages,
    include_package_data=True,
    install_requires=[f"avocado-framework=={VERSION}"],
    test_suite="tests",
    entry_points={
        "avocado.plugins.cli": [
            "varianter_cit = avocado_varianter_cit.varianter_cit:VarianterCitCLI",
        ],
        "avocado.plugins.varianter": [
            "varianter_cit = avocado_varianter_cit.varianter_cit:VarianterCit",
        ],
    },
)



import itertools


class CombinationRow:

    """
    Row object store all combinations between two parameters into dictionary.
    Keys in dictionary are values of combinations and values in dictionary are
    information about coverage. Row object has information how many combinations
    are uncovered and how many of them are covered more than ones.
    """

    def __init__(self, input_data, t_value, parameters):
        """
        :param input_data: list of data from user
        :param t_value: t number from user
        :param parameters: the tuple of parameters whose combinations Row object represents
        """

        self.hash_table = {}
        self.covered_more_than_ones = 0
        self.uncovered = 0
        array = []
        "Creation of combinations"
        for i in range(t_value):
            array.append(list(range(input_data[parameters[i]])))
        for i in itertools.product(*array):
            self.uncovered += 1
            self.hash_table[i] = 0

    def cover_cell(self, key):
        """
        Cover one combination inside Row

        :param key: combination to be covered
        :return: number of new covered combinations and number of new covered combinations more than ones
        """

        old_uncovered = self.uncovered
        old_covered_more_than_ones = self.covered_more_than_ones
        value = self.hash_table[tuple(key)]
        if value is not None:
            if value == 0:
                self.uncovered -= 1
            elif value == 1:
                self.covered_more_than_ones += 1
            self.hash_table[tuple(key)] += 1

        return (
            self.uncovered - old_uncovered,
            self.covered_more_than_ones - old_covered_more_than_ones,
        )

    def uncover_cell(self, key):
        """
        Uncover one combination inside Row

        :param key: combination to be uncovered
        :return: number of new covered combinations and number of new covered combinations more than ones
        """

        old_uncovered = self.uncovered
        old_covered_more_than_ones = self.covered_more_than_ones
        value = self.hash_table[tuple(key)]
        if value is not None and value > 0:
            if value == 1:
                self.uncovered += 1
            elif value == 2:
                self.covered_more_than_ones -= 1
            self.hash_table[tuple(key)] -= 1

        return (
            self.uncovered - old_uncovered,
            self.covered_more_than_ones - old_covered_more_than_ones,
        )

    def completely_uncover(self):
        """
        Uncover all combinations inside Row
        """

        self.uncovered = 0
        self.covered_more_than_ones = 0
        for key, value in self.hash_table.items():
            if value is not None:
                self.hash_table[key] = 0
                self.uncovered += 1

    def del_cell(self, key):
        """
        Disable one combination. If combination is disabled it means that
        the combination does not match the constraints

        :param key: combination to be disabled
        :return: number of new covered combinations
        """

        key = tuple(key)
        if self.hash_table[key] is not None:
            self.hash_table[key] = None
            self.uncovered -= 1
            return -1
        else:
            return 0

    def is_valid(self, key):
        """
        Is the combination match the constraints.

        :param key: combination to valid
        """

        key = tuple(key)
        if self.hash_table.get(key, 0) is None:
            return False
        else:
            return True

    def get_all_uncovered_combinations(self):
        """
        :return: list of all uncovered combination
        """

        combinations = []
        for key, value in self.hash_table.items():
            if value == 0:
                combinations.append(key)
        return combinations

    def __eq__(self, other):
        return (
            self.covered_more_than_ones == other.covered_more_than_ones
            and self.uncovered == other.uncovered
            and self.hash_table == other.hash_table
        )
